/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: 11 Steps

   Minimalistic examples to build upon and experiment with 
	 to develop an understanding of the underlying API

	   
/*
/****************************************************************/
/*
     About:
	 Self contained demo (i.e., all the code is in main.cpp) - 
	 however!  You also need to include the two binary shader
	 files (vertex and shader)

	 ''''
	 Modifications to the original book, code is modularized
	 into functions (i.e., shows what parameters go in/out)
	 and additional depth-buffer define (turn on/off) 
	 to include zbuffer code.
	 ''''

*/


#include "setup.h"


int CALLBACK WinMain( HINSTANCE hInstance, 
					  HINSTANCE hPrevInstance, 
					  LPSTR		lpCmdLine, 
					  int		nCmdShow ) 
{
	// Step-1-
	int				width				= 800;
	int				height				= 600;
	HWND			windowHandle		= NULL;
	SetupWindow(width, height, &windowHandle);

	// Step-2-
	VkInstance		instance			= NULL;
	VkSurfaceKHR	surface				= NULL;
	SetupVulkanInstance(windowHandle,
		                &instance,
					    &surface);

	// Step-3-
	VkPhysicalDevice physicalDevice		= NULL;
  VkDevice	device					        = NULL;
	SetupPhysicalDevice(instance,
                        &physicalDevice,
                        &device);

	// Step-4-
	VkSwapchainKHR   swapChain			= NULL;					     
	VkImage*         presentImages		= NULL;					     
	VkImageView*     presentImageViews	= NULL;
	SetupSwapChain( device,
					physicalDevice,
					surface,
					&width,
					&height,
					&swapChain,			     
					&presentImages,				     
					&presentImageViews);

  // Step-5-
	VkRenderPass     renderPass			= NULL;
	VkFramebuffer*   frameBuffers		= NULL;
	SetupRenderPass(device,
					physicalDevice,
					width,
					height,
					presentImageViews,
					&renderPass, 
					&frameBuffers);

	// Step-6-
	VkCommandBuffer  commandBuffer      = NULL;
  VkCommandPool    commandPool        = NULL;
	SetupCommandBuffer(device,
					   physicalDevice,
             &commandPool,
					   &commandBuffer);
	

	// Step-7-
	VkBuffer         vertexInputBuffer = NULL;
	int				 vertexSize		   = 0;
	int				 numberOfTriangles = 0;
	SetupVertexBuffer(device,
					  physicalDevice,
					  &vertexSize,
					  &numberOfTriangles,
					  &vertexInputBuffer );

	// Step-8-
	VkShaderModule   vertShaderModule  = NULL;
	VkShaderModule   fragShaderModule  = NULL;
	VkBuffer		 buffer                = NULL;
	VkDeviceMemory	 memory            = NULL;
	SetupShaderandUniforms(device,
						   physicalDevice,
						   width,
						   height,
						   &vertShaderModule,
						   &fragShaderModule,
						   &buffer,
						   &memory);

	// Step-9-
	VkDescriptorSet       descriptorSet			= NULL;
	VkDescriptorSetLayout descriptorSetLayout   = NULL;
	SetupDescriptors(device,
                   physicalDevice,
                   commandPool,
					         buffer,
					         &descriptorSet,
					         &descriptorSetLayout);

	// Step-10-
	VkPipeline       pipeline          = NULL;
	VkPipelineLayout pipelineLayout    = NULL;
	SetupPipeline(	device,
				    width,
					height,
					vertexSize,
					descriptorSetLayout,
					vertShaderModule,
					fragShaderModule,
					renderPass, 
					&pipeline,
					&pipelineLayout);



  // Step-11-
	MSG msg;
	while( true ) 
	{
		// Continually force the window to be
		// redrawn as long as no other Win32 
		// messages are pending
		PeekMessage( &msg, NULL, NULL, NULL, PM_REMOVE );
		if( msg.message == WM_QUIT ) 
		{
			break;
		} 
		else 
		{
			TranslateMessage( &msg ); 
			DispatchMessage( &msg );
		}

		RenderLoop( device,
					width,
					height,
					numberOfTriangles,
					swapChain,
					commandBuffer,
					presentImages,
					frameBuffers,
					renderPass,
					vertexInputBuffer,
					descriptorSet,
					pipelineLayout,
					pipeline);
	}
	return 0;
}// End WinMain(..)

