/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: 11 Steps

     Minimalistic examples to build upon and experiment with 
	 to develop an understanding of the underlying API

	   
/*
/****************************************************************/

// Fragement Shader or sometimes called the 
// `pixel shader'
// shader version
#version 420

// input (from our vertex shader)
layout ( location = 0 ) in vec3 inEyeSpace;
layout ( location = 1 ) in vec4 inColor;
layout ( location = 2 ) in vec2 inTexCoords;
layout ( location = 3 ) in vec3 inNormal;

layout ( binding = 1 ) uniform samplerCube inTexCubeSampler;

// final screen output color
layout ( location = 0 ) out vec4 outFragColor;

// define if you want the fog as a `plane' or as a `dome'
int depthFog    = 0;
// type of fog, linear, exp, ..
int fogSelector = 0;

// shader entry point
void main() 
{
  // distance
  float dist      = 0;
  float fogFactor = 0;
  
  // compute distance used in fog equations
  if ( depthFog == 0 ) // select plane or range based
  {
    // plane based
    dist = abs(inEyeSpace.z);
  }
  else
  {
     //range based
     dist = length(inEyeSpace);
  }
  

  if ( fogSelector == 0 ) // linear fog
  {
     // Fog `effect' doesn't influence the lighting until
	 // it has passed the distance of `10' units
	 // Gradually increases until it reaches 50 units
	 // then it's fully fogged!

     // 10 - fog starts; 50 - fog ends
     fogFactor = (50 - dist)/(50 - 10);
     fogFactor = 1.0 - clamp( fogFactor, 0.0, 1.0 );
  }
  else // exponential fog
  {
      float fogDensity = 1.5;
  	fogFactor = 1.0 /exp(dist * fogDensity);
  	fogFactor = 1.0 - clamp( fogFactor, 0.0, 1.0 );
  }

  // fogFactor:
  // 0 - original color
  // 1 - fog color
  outFragColor = mix( inColor, vec4(0.2,0.2,0.2,1.0), fogFactor );

}// End main(..)