/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: 11 Steps

     Minimalistic examples to build upon and experiment with 
	 to develop an understanding of the underlying API
	 
	 Vertex Shader
/*
/****************************************************************/


// shader version
#version 420


// input matrix transforms (3D) passed in from our program
layout ( std140, set = 0, binding = 0 ) uniform buffer 
{
  mat4 inProjectionMatrix;
  mat4 inViewMatrix;
  mat4 inModelMatrix;
};

// input from your program
layout( location = 0 ) in vec4 inPos;
layout( location = 1 ) in vec3 inColor;
layout( location = 2 ) in vec2 inTexCoords;
layout( location = 3 ) in vec3 inNormal;

// output to fragment shader
layout( location = 0 ) out vec3 outEyeSpace;
layout( location = 1 ) out vec4 outColor;
layout( location = 2 ) out vec2 outTexCoords;
layout( location = 3 ) out vec3 outNormal;

// shader entry point
void main() 
{	
  // combine the matrices
  mat4 mvp = inModelMatrix * inViewMatrix * inProjectionMatrix;

  // transform position by matrices
  gl_Position = vec4(inPos.xyz, 1) * mvp;

  //send it to fragment shader
  vec4 eyeSpace = vec4(inPos.xyz,1) * (inModelMatrix * inViewMatrix);
  outEyeSpace = eyeSpace.xyz / eyeSpace.w;

  // pass onto the next stage
  outColor     = vec4(inColor, 1.0);
  outTexCoords = inTexCoords;
  
  // pass input normals onto the fragment shader
  outNormal = mat3(transpose(inverse(inModelMatrix))) * inNormal;

}// End main(..)