
#pragma once


#if 0

// Vulkan defines, e.g., VkSampler, VkImage, ..
#include "setup.h"

#include "loadtga.h"

/**************************************************************/
/*
   Create/setup and contain the texture information
   Currently uses a simple test pattern, however, can be 
   modified with a few lines of code to load a texture

   Class name: Texture
   File: texture.h
*/
/**************************************************************/

class Texture
{
public:

  VkSampler              sampler               ;//= NULL;
  VkImage                image                 ;//= NULL;
  VkImageView            imageView             ;//= NULL;
  VkDeviceMemory         imageMemory           ;//= { 0 };
  VkDescriptorImageInfo  descriptorImageInfo;
  int                    width                 ;//= 0;
  int                    height                ;//= 0;

  Texture(VkDevice         device,
          VkPhysicalDevice physicalDevice,
          const char*      tgaFileName)
  {
    sampler               = NULL;
    image                 = NULL;
    imageView             = NULL;
    imageMemory           = 0;
    descriptorImageInfo;
    width                 = 0;
    height                = 0;

    VkSamplerCreateInfo samplerCreateInfo      = {};
    samplerCreateInfo.sType                    = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerCreateInfo.magFilter                = VK_FILTER_LINEAR;
    samplerCreateInfo.minFilter                = VK_FILTER_LINEAR;
    samplerCreateInfo.mipmapMode               = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    samplerCreateInfo.addressModeU             = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.addressModeV             = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.addressModeW             = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.mipLodBias               = 0;
    samplerCreateInfo.anisotropyEnable         = VK_FALSE;
    samplerCreateInfo.minLod                   = 0;
    samplerCreateInfo.maxLod                   = 5;
    samplerCreateInfo.borderColor              = VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
    samplerCreateInfo.unnormalizedCoordinates  = VK_FALSE;

    
    VkResult result = 
    vkCreateSampler( device, 
                     &samplerCreateInfo, 
                     NULL, 
                     &sampler );

    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to create sampler." );


#if 0 // Testing using a procedurally generated pattern

	// Create a test image 512x512 - you could modify this
	// so the image is loaded in from a file (e.g., tga loader)
	// for testing - you'll just create a pattern
    width  = 512;
    height = 512;
    void* imageData   = (void *) new DWORD[ width * height ];

  
    // generate a pretty test pattern
    for (int y=0; y<height; ++y)
    {
      for (int x=0; x<width; ++x)
      {
        int cc = (x^y)   &0xFF;
        int r  = (cc>>16)&0xFF;
        int g  = (cc>>8) &0xFF;
        int b  = (cc>>0) &0xFF ;
        // beautiful xor pattern
        DWORD* pixel = ((DWORD *) imageData) + ( x * height  ) + ( y );
        pixel[0] = cc;
      }// End inner loop
    }// End outer loop
#else
    TGAFILE tgaFile;
    bool successful =
    LoadTGA(tgaFileName, &tgaFile);

    DBG_ASSERT(true == successful);

    width  = tgaFile.imageWidth;
    height = tgaFile.imageHeight;

    // Simple sanity checks
    DBG_ASSERT(width >0 && width <99999);
    DBG_ASSERT(height>0 && height<99999);

    // Only support 32 bit tga
    // i.e., 4 bytes per pixel color
    DBG_ASSERT(tgaFile.bitCount == 32 );

    void* imageData   = tgaFile.imageData;
    DBG_ASSERT(imageData);

#endif

    VkImageCreateInfo textureCreateInfo        = {};
    textureCreateInfo.sType                    = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    textureCreateInfo.imageType                = VK_IMAGE_TYPE_2D;
    textureCreateInfo.format                   = VK_FORMAT_R8G8B8A8_UNORM;;
    textureCreateInfo.extent.width             = width; 
    textureCreateInfo.extent.height            = height; 
    textureCreateInfo.extent.depth             = 1; 
    textureCreateInfo.mipLevels                = 1;
    textureCreateInfo.arrayLayers              = 1;
    textureCreateInfo.samples                  = VK_SAMPLE_COUNT_1_BIT;
    textureCreateInfo.tiling                   = VK_IMAGE_TILING_LINEAR;
    textureCreateInfo.usage                    = VK_IMAGE_USAGE_SAMPLED_BIT;
    textureCreateInfo.sharingMode              = VK_SHARING_MODE_EXCLUSIVE;
    textureCreateInfo.initialLayout            = VK_IMAGE_LAYOUT_PREINITIALIZED;

    result = 
    vkCreateImage( device, 
             &textureCreateInfo, 
             NULL, 
             &image );

    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to create texture image." );

    // allocate and bind image memory:
    VkMemoryRequirements textureMemoryRequirements = {};
    vkGetImageMemoryRequirements( device, 
                    image, 
                    &textureMemoryRequirements );

    VkMemoryAllocateInfo textureImageAllocateInfo = {};
    textureImageAllocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    textureImageAllocateInfo.allocationSize = textureMemoryRequirements.size;

    // read the device memory properties
    VkPhysicalDeviceMemoryProperties memoryProperties;
    vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

    uint32_t textureMemoryTypeBits = textureMemoryRequirements.memoryTypeBits;
    VkMemoryPropertyFlags tDesiredMemoryFlags = VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT;
    for( uint32_t i = 0; i < 32; ++i ) 
    {
      VkMemoryType memoryType = memoryProperties.memoryTypes[i];
      if( textureMemoryTypeBits & 1 ) 
      {
        if( ( memoryType.propertyFlags & tDesiredMemoryFlags ) == tDesiredMemoryFlags ) 
        {
          textureImageAllocateInfo.memoryTypeIndex = i;
          break;
        }
      }
      textureMemoryTypeBits = textureMemoryTypeBits >> 1;
    }// End for i

    result = vkAllocateMemory( device, &textureImageAllocateInfo, NULL, &imageMemory );
    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to allocate device memory." );

    result = vkBindImageMemory( device, image, imageMemory, 0 );
    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to bind image memory." );

    // copy our image content:
    void *imageMapped;
    result = vkMapMemory( device, imageMemory, 0, VK_WHOLE_SIZE, 0, &imageMapped );
    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to map uniform buffer memory." );
    
    memcpy( imageMapped, imageData, sizeof(DWORD) * width * height  );

    VkMappedMemoryRange memoryRange = {};
    memoryRange.sType               = VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE;
    memoryRange.memory              = imageMemory;
    memoryRange.offset              = 0;
    memoryRange.size                = VK_WHOLE_SIZE;
    vkFlushMappedMemoryRanges( device, 1, &memoryRange );
    
    vkUnmapMemory( device, imageMemory );

  
    delete[] imageData;
    imageData = NULL;

    // create and map the texture view
    VkImageViewCreateInfo textureImageViewCreateInfo           = {};
    textureImageViewCreateInfo.sType                           = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    textureImageViewCreateInfo.image                           = image;
    textureImageViewCreateInfo.viewType                        = VK_IMAGE_VIEW_TYPE_2D;
    textureImageViewCreateInfo.format                          = VK_FORMAT_R8G8B8A8_UNORM;
    VkComponentMapping j = { VK_COMPONENT_SWIZZLE_R, 
                             VK_COMPONENT_SWIZZLE_G, 
                             VK_COMPONENT_SWIZZLE_B, 
                             VK_COMPONENT_SWIZZLE_A };
    textureImageViewCreateInfo.components                      = j; 
    textureImageViewCreateInfo.subresourceRange.aspectMask     = VK_IMAGE_ASPECT_COLOR_BIT;
    textureImageViewCreateInfo.subresourceRange.baseMipLevel   = 0;
    textureImageViewCreateInfo.subresourceRange.levelCount     = 1;
    textureImageViewCreateInfo.subresourceRange.baseArrayLayer = 0;
    textureImageViewCreateInfo.subresourceRange.layerCount     = 1;

    result = 
    vkCreateImageView( device, 
               &textureImageViewCreateInfo, 
               NULL, 
               &imageView );

    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to create image view." );

    // Fill descriptor image info that can be used for setting up descriptor sets
    descriptorImageInfo.imageLayout = VK_IMAGE_LAYOUT_GENERAL;
    descriptorImageInfo.imageView   = imageView;
    descriptorImageInfo.sampler     = sampler;
  }// End Texture(..) constructor

};// End class Texture





#endif

