

#pragma once


/****************************************************************/
/*
   Create/setup and contain a cube texture 
   Currently uses six seperate images and builds up the cube
   map texture 

   Class name: CubeTexture
   File: cubetexture.h

   Cube texture loading class

*/
/****************************************************************/

// Vulkan defines, e.g., VkSampler, VkImage, ..
#include "setup.h"

#include "loadtga.h"



#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>



class TextureCube
{
public:
  VkSampler              sampler              ;//= NULL;
  VkImage                image                ;//= NULL;
  VkImageView            imageView            ;//= NULL;
  VkDeviceMemory         imageMemory          ;//= { 0 };
  VkDescriptorImageInfo  descriptorImageInfo;
  int                    width                ;//= 0;
  int                    height               ;//= 0;
  int                    numCubeSides         ;//= 6;


  TextureCube(VkDevice         device,
              VkPhysicalDevice physicalDevice,
              VkCommandPool    cmdPool,
              const char**     arrayTgaFileName)
  {
    /*
       For example:
       const char* arrayTgaFileName[] = {
        "assets/rockyvalley_rt.tga",
        "assets/rockyvalley_lf.tga",
        "assets/rockyvalley_up.tga",
        "assets/rockyvalley_dn.tga",
        "assets/rockyvalley_bk.tga",
        "assets/rockyvalley_ft.tga",
        };
    */

    sampler                                   = NULL;
    image                                     = NULL;
    imageView                                 = NULL;
    imageMemory                               = 0;
    width                                     = 0;
    height                                    = 0;
    numCubeSides                              = 6;

    // Temp - hard code width/height of the cube map images
    width                                     = 512;
    height                                    = 512;

    VkSamplerCreateInfo samplerCreateInfo     = {};
    samplerCreateInfo.sType                   = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
    samplerCreateInfo.magFilter               = VK_FILTER_LINEAR;
    samplerCreateInfo.minFilter               = VK_FILTER_LINEAR;
    samplerCreateInfo.mipmapMode              = VK_SAMPLER_MIPMAP_MODE_LINEAR;
    samplerCreateInfo.addressModeU            = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.addressModeV            = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.addressModeW            = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
    samplerCreateInfo.mipLodBias              = 0;
    samplerCreateInfo.anisotropyEnable        = VK_FALSE;
    samplerCreateInfo.minLod                  = 0;
    samplerCreateInfo.maxLod                  = 5;
    samplerCreateInfo.borderColor             = VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
    samplerCreateInfo.unnormalizedCoordinates = VK_FALSE;
    samplerCreateInfo.anisotropyEnable        = VK_TRUE;

    
    VkResult result = 
    vkCreateSampler( device, 
                     &samplerCreateInfo, 
                     NULL, 
                     &sampler );

    DBG_ASSERT_VULKAN_MSG( result, 
      "Failed to create sampler." );





    VkBufferCreateInfo bufferCreateInfo;
    memset(&bufferCreateInfo, 0, sizeof(bufferCreateInfo));
    bufferCreateInfo.sType                    = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
    bufferCreateInfo.size                     = 512 * 512 * 4 * numCubeSides; // 4->RGBA
    // This buffer is used as a transfer source for the buffer copy
    bufferCreateInfo.usage                    = VK_BUFFER_USAGE_TRANSFER_SRC_BIT;
    bufferCreateInfo.sharingMode              = VK_SHARING_MODE_EXCLUSIVE;

#if 1
    // Create a host-visible staging buffer that contains the raw image data
    VkBuffer stagingBuffer;
    DBG_ASSERT_VULKAN(vkCreateBuffer(device, &bufferCreateInfo, NULL, &stagingBuffer));

    // Get memory requirements for the staging buffer (alignment, memory type bits)
    VkMemoryAllocateInfo memAllocInfo;
    memset(&memAllocInfo, 0, sizeof(memAllocInfo));
    memAllocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
    VkMemoryRequirements memReqs;
    
    vkGetBufferMemoryRequirements(device, stagingBuffer, &memReqs);
    memAllocInfo.allocationSize = memReqs.size;
#endif



    unsigned char* imageData   = (unsigned char *) new DWORD[ width * height * numCubeSides ];

    /*
    const char* tgafiles[] = {
    "assets/rockyvalley_rt.tga",
    "assets/rockyvalley_lf.tga",
    "assets/rockyvalley_up.tga",
    "assets/rockyvalley_dn.tga",
    "assets/rockyvalley_bk.tga",
    "assets/rockyvalley_ft.tga",
    };
    */

    int offs = 0;
    for (int i=0; i<numCubeSides; ++i)
    {
      TGAFILE tgaFile;
      bool err =
      LoadTGA( arrayTgaFileName[i], &tgaFile );
      DBG_ASSERT(err);


      DBG_ASSERT(tgaFile.imageWidth == 512 );
      DBG_ASSERT(tgaFile.imageHeight == 512 );

      DBG_ASSERT(tgaFile.bitCount == 32 );
      unsigned char* imageDataTGA = (unsigned char*)tgaFile.imageData;

      memcpy( imageData+offs, imageDataTGA, 512*512*4 );
      offs += 512*512*4;
    }

#if 1
    memAllocInfo.memoryTypeIndex = 
      GetMemoryType(physicalDevice,
                    memReqs.memoryTypeBits, 
                    VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT);

    VkDeviceMemory stagingMemory;
    DBG_ASSERT_VULKAN(vkAllocateMemory(device, &memAllocInfo, nullptr, &stagingMemory));
    DBG_ASSERT_VULKAN(vkBindBufferMemory(device, stagingBuffer, stagingMemory, 0));

    // Copy texture data into staging buffer
    uint8_t *data;
    DBG_ASSERT_VULKAN(vkMapMemory(device, stagingMemory, 0, memReqs.size, 0, (void **)&data));
    memcpy(data, imageData, 512*512*4 * numCubeSides);
    vkUnmapMemory(device, stagingMemory);

    // Create optimal target image
    VkImageCreateInfo imageCreateInfo;
    memset(&imageCreateInfo, 0, sizeof(imageCreateInfo));
    imageCreateInfo.sType              = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
    imageCreateInfo.imageType          = VK_IMAGE_TYPE_2D;
    imageCreateInfo.format             = VK_FORMAT_R8G8B8A8_UNORM;
    imageCreateInfo.mipLevels          = 1;
    imageCreateInfo.samples            = VK_SAMPLE_COUNT_1_BIT;
    imageCreateInfo.tiling             = VK_IMAGE_TILING_OPTIMAL;
    imageCreateInfo.usage              = VK_IMAGE_USAGE_SAMPLED_BIT;
    imageCreateInfo.sharingMode        = VK_SHARING_MODE_EXCLUSIVE;
    imageCreateInfo.initialLayout      = VK_IMAGE_LAYOUT_UNDEFINED;
    const VkExtent3D ex = {512, 512, 1};
    imageCreateInfo.extent             = ex;
    imageCreateInfo.usage              = VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT;
    // Cube faces count as array layers in Vulkan
    imageCreateInfo.arrayLayers        = numCubeSides;
    // This flag is required for cube map images
    imageCreateInfo.flags              = VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT;

    DBG_ASSERT_VULKAN(vkCreateImage(device, &imageCreateInfo, nullptr, &image));

    vkGetImageMemoryRequirements(device, image, &memReqs);

    memAllocInfo.allocationSize = memReqs.size;
    memAllocInfo.memoryTypeIndex = GetMemoryType(physicalDevice,
                                                 memReqs.memoryTypeBits, 
                                                 VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);

    DBG_ASSERT_VULKAN(vkAllocateMemory(device, &memAllocInfo, nullptr, &imageMemory));
    DBG_ASSERT_VULKAN(vkBindImageMemory(device, image, imageMemory, 0));

    //==============================

    VkCommandBuffer copyCmd = CreateCommandBuffer(device, cmdPool);

    VkCommandBufferBeginInfo cmdBufferBeginInfo;
    memset(&cmdBufferBeginInfo, 0, sizeof(cmdBufferBeginInfo));
    cmdBufferBeginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;

    DBG_ASSERT_VULKAN(vkBeginCommandBuffer(copyCmd, &cmdBufferBeginInfo));


    // Setup buffer copy regions for each face including all of it's miplevels
    std::vector<VkBufferImageCopy> bufferCopyRegions;
    uint32_t offset = 0;

    for (uint32_t face = 0; face < (uint32_t)numCubeSides; ++face)
    {
      VkBufferImageCopy bufferCopyRegion = {};
      bufferCopyRegion.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
      bufferCopyRegion.imageSubresource.mipLevel = 0;
      bufferCopyRegion.imageSubresource.baseArrayLayer = face;
      bufferCopyRegion.imageSubresource.layerCount = 1;
      bufferCopyRegion.imageExtent.width =  512;
      bufferCopyRegion.imageExtent.height = 512;
      bufferCopyRegion.imageExtent.depth = 1;
      bufferCopyRegion.bufferOffset = offset;

      bufferCopyRegions.push_back(bufferCopyRegion);

      // Increase offset into staging buffer for next face
      offset += 512*512*4;
    }// End for face

    // Image barrier for optimal image (target)
    // Set initial layout for all array layers (faces) of the optimal (target) tiled texture
    VkImageSubresourceRange subresourceRange = {};
    subresourceRange.aspectMask              = VK_IMAGE_ASPECT_COLOR_BIT;
    subresourceRange.baseMipLevel            = 0;
    subresourceRange.levelCount              = 1;
    subresourceRange.layerCount              = numCubeSides;

    
    SetImageLayout(
      copyCmd,
      image,
      VK_IMAGE_ASPECT_COLOR_BIT,
      VK_IMAGE_LAYOUT_UNDEFINED,
      VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
      subresourceRange);

    // Copy the cube map faces from the staging buffer to the optimal tiled image
    vkCmdCopyBufferToImage(
      copyCmd,
      stagingBuffer,
      image,
      VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
      static_cast<uint32_t>(bufferCopyRegions.size()),
      bufferCopyRegions.data()
      );

    // Change texture image layout to shader read after all faces have been copied
    VkImageLayout imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
    SetImageLayout(
      copyCmd,
      image,
      VK_IMAGE_ASPECT_COLOR_BIT,
      VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
      imageLayout,
      subresourceRange);

    // Flush
    {
      DBG_ASSERT_VULKAN(vkEndCommandBuffer(copyCmd));

      VkSubmitInfo submitInfo = {};
      submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
      submitInfo.commandBufferCount = 1;
      submitInfo.pCommandBuffers = &copyCmd;

      VkQueue queue = NULL;
      vkGetDeviceQueue( device, 0, 0, &queue );
      
      DBG_ASSERT_VULKAN(vkQueueSubmit(queue, 1, &submitInfo, VK_NULL_HANDLE));
      DBG_ASSERT_VULKAN(vkQueueWaitIdle(queue));

      vkFreeCommandBuffers(device, cmdPool, 1, &copyCmd);
    }
#endif

    // Create image view
    VkImageViewCreateInfo view;
    memset(&view, 0, sizeof(view));
    view.sType                       = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
    // Cube map view type
    view.viewType                    = VK_IMAGE_VIEW_TYPE_CUBE;
    view.format                      = VK_FORMAT_R8G8B8A8_UNORM;
    const
    VkComponentMapping j = { VK_COMPONENT_SWIZZLE_R, 
                             VK_COMPONENT_SWIZZLE_G, 
                             VK_COMPONENT_SWIZZLE_B, 
                             VK_COMPONENT_SWIZZLE_A };
    view.components                  = j;
    const
    VkImageSubresourceRange k = { VK_IMAGE_ASPECT_COLOR_BIT, 0, 1, 0, 1 };
    view.subresourceRange            = k;
    // 6 array layers (faces)
    view.subresourceRange.layerCount = numCubeSides;
    // Set number of mip levels
    view.subresourceRange.levelCount = 1;
    view.image                       = image;
    DBG_ASSERT_VULKAN(vkCreateImageView(device, &view, nullptr,  &imageView));

#if 1
    // Clean up staging resources
    vkFreeMemory(device, stagingMemory, nullptr);
    vkDestroyBuffer(device, stagingBuffer, nullptr);
#endif

    // Fill descriptor image info that can be used for setting up descriptor sets
    descriptorImageInfo.imageLayout = VK_IMAGE_LAYOUT_GENERAL;
    descriptorImageInfo.imageView   = imageView;
    descriptorImageInfo.sampler     = sampler;
  }



  uint32_t GetMemoryType(VkPhysicalDevice       physicalDevice,
                         uint32_t               typeBits, 
                         VkMemoryPropertyFlags  properties, 
                         VkBool32 *             memTypeFound = nullptr)
  {
    // read the device memory properties
    VkPhysicalDeviceMemoryProperties memoryProperties;
    vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

    for (uint32_t i = 0; i < memoryProperties.memoryTypeCount; i++)
    {
      if ((typeBits & 1) == 1)
      {
        if ((memoryProperties.memoryTypes[i].propertyFlags & properties) == properties)
        {
          if (memTypeFound)
          {
            *memTypeFound = true;
          }
          return i;
        }
      }
      typeBits >>= 1;
    }// End for i
    DBG_ASSERT(false);
    return 0;
  }// End GetMemoryType(..)


  void SetImageLayout(VkCommandBuffer cmdbuffer, 
                      VkImage image, 
                      VkImageAspectFlags aspectMask, 
                      VkImageLayout oldImageLayout, 
                      VkImageLayout newImageLayout,
                      VkImageSubresourceRange subresourceRange,
                      VkPipelineStageFlags srcStageMask = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT,
                      VkPipelineStageFlags dstStageMask = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT)
  {
    // Create an image barrier object
    VkImageMemoryBarrier imageMemoryBarrier;
    memset(&imageMemoryBarrier, 0, sizeof(imageMemoryBarrier));
    imageMemoryBarrier.sType               = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
    imageMemoryBarrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    imageMemoryBarrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
    imageMemoryBarrier.oldLayout           = oldImageLayout;
    imageMemoryBarrier.newLayout           = newImageLayout;
    imageMemoryBarrier.image               = image;
    imageMemoryBarrier.subresourceRange    = subresourceRange;

    // Source layouts (old)
    // Source access mask controls actions that have to be finished on the old layout
    // before it will be transitioned to the new layout
    switch (oldImageLayout)
    {
    case VK_IMAGE_LAYOUT_UNDEFINED:
        // Image layout is undefined (or does not matter)
        // Only valid as initial layout
        // No flags required, listed only for completeness
        imageMemoryBarrier.srcAccessMask = 0;
        break;

    case VK_IMAGE_LAYOUT_PREINITIALIZED:
        // Image is preinitialized
        // Only valid as initial layout for linear images, preserves memory contents
        // Make sure host writes have been finished
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_HOST_WRITE_BIT;
        break;

    case VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL:
        // Image is a color attachment
        // Make sure any writes to the color buffer have been finished
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
        break;

    case VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL:
        // Image is a depth/stencil attachment
        // Make sure any writes to the depth/stencil buffer have been finished
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT;
        break;

    case VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL:
        // Image is a transfer source 
        // Make sure any reads from the image have been finished
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
        break;

    case VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL:
        // Image is a transfer destination
        // Make sure any writes to the image have been finished
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        break;

    case VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL:
        // Image is read by a shader
        // Make sure any shader reads from the image have been finished
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_SHADER_READ_BIT;
        break;
    }

    // Target layouts (new)
    // Destination access mask controls the dependency for the new image layout
    switch (newImageLayout)
    {
    case VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL:
      // Image will be used as a transfer destination
      // Make sure any writes to the image have been finished
      imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
      break;

    case VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL:
      // Image will be used as a transfer source
      // Make sure any reads from and writes to the image have been finished
      imageMemoryBarrier.srcAccessMask = imageMemoryBarrier.srcAccessMask | VK_ACCESS_TRANSFER_READ_BIT;
      imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
      break;

    case VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL:
      // Image will be used as a color attachment
      // Make sure any writes to the color buffer have been finished
      imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_READ_BIT;
      imageMemoryBarrier.dstAccessMask = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
      break;

    case VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL:
      // Image layout will be used as a depth/stencil attachment
      // Make sure any writes to depth/stencil buffer have been finished
      imageMemoryBarrier.dstAccessMask = imageMemoryBarrier.dstAccessMask | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT;
      break;

    case VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL:
      // Image will be read in a shader (sampler, input attachment)
      // Make sure any writes to the image have been finished
      if (imageMemoryBarrier.srcAccessMask == 0)
      {
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_HOST_WRITE_BIT | VK_ACCESS_TRANSFER_WRITE_BIT;
      }
      imageMemoryBarrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;
      break;
    }

    // Put barrier inside setup command buffer
    vkCmdPipelineBarrier(
      cmdbuffer, 
      srcStageMask, 
      dstStageMask, 
      0, 
      0, nullptr,
      0, nullptr,
      1, &imageMemoryBarrier);
  }

  VkCommandBuffer CreateCommandBuffer(VkDevice         device,
                                      VkCommandPool    cmdPool)
  {
    // Create one command buffer for each swap chain image and reuse for rendering
    VkCommandBuffer cmdBuffer;

    VkCommandBufferAllocateInfo cmdBufAllocateInfo = {};
    cmdBufAllocateInfo.sType                       = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
    cmdBufAllocateInfo.commandPool                 = cmdPool;
    cmdBufAllocateInfo.level                       = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
    cmdBufAllocateInfo.commandBufferCount          = 1;

    DBG_ASSERT_VULKAN(vkAllocateCommandBuffers(device, &cmdBufAllocateInfo, &cmdBuffer));

    return cmdBuffer;
  }
};